import os

import npc_lims
import pandas as pd

import npc_sessions.nwb as nwb


def run_capsules_for_units_kilosort_codeocean(session_id: str) -> None:
    raw_data_asset = npc_lims.get_session_raw_data_asset(session_id)
    if raw_data_asset is None:
        raise ValueError(
            f"Session {session_id} has no raw data asset."
        )  # TODO move to function in npc_lims
    sorted_data_asset = npc_lims.get_session_sorted_data_asset(session_id)
    if sorted_data_asset is None:
        raise ValueError(
            f"Session {session_id} has no sorted data asset."
        )  # TODO move to function in npc_lims
    # capsule_results_units = npc_lims.run_capsule_and_get_results('980c5218-abef-41d8-99ed-24798d42313b',
    # (raw_data_asset, sorted_data_asset))
    # npc_lims.register_session_data_asset(self.id, capsule_results_units)

    units_no_peak_channel_asset = npc_lims.get_session_units_data_asset(session_id)
    if units_no_peak_channel_asset is None:
        raise ValueError(
            f"Session {session_id} has no peak_channel data asset."
        )  # TODO move to function in npc_lims

    capsule_result_units_peak_channels = npc_lims.run_capsule_and_get_results(
        "d1a5c3a8-8fb2-4cb0-8e9e-96e6e1d03ff1",
        (raw_data_asset, sorted_data_asset, units_no_peak_channel_asset),
    )
    npc_lims.register_session_data_asset(session_id, capsule_result_units_peak_channels)


def get_units(
    session_id: str, method: str = "kilosort_codeocean"
) -> tuple[nwb.Units, ...]:
    records = npc_lims.NWBSqliteDBHub().get_records(
        npc_lims.Units, session_id=session_id
    )
    if not records:
        units_s3_path = npc_lims.get_units_file_from_s3(session_id)

        if units_s3_path is None:
            # TODO: DO SOMETHING TO GET UNITS - Right now, run spike nwb capsule, and register units as data asset
            if method == "kilosort_codeocean":
                run_capsules_for_units_kilosort_codeocean(session_id)
                units_s3_path = npc_lims.get_units_file_from_s3(session_id)

        if units_s3_path is not None:  # TODO: change to polars
            pd.read_csv(
                units_s3_path,
                storage_options={
                    "key": os.getenv("AWS_ACCESS_KEY_ID"),
                    "secret": os.getenv("AWS_SECRET_ACCESS_KEY"),
                },
            )
            print()

    return ()



"""
getting Optotagging trials table requires:
- one or more hdf5 files with trial/stim data, called 'OptoTagging_*.hdf5'
- frame display times from sync, to use in place of frametimes in hdf5 file
- latency estimate for each stim presentation, to be added to frame
display times to get stim onset times
"""
import contextlib
import functools

import numpy as np
import numpy.typing as npt

import npc_sessions.utils as utils
from npc_sessions.trials.TaskControl import TaskControl


class OptoTagging(TaskControl):
    _stim_onset_times: npt.NDArray[np.float64]
    """`[1 x num trials]` onset time of each opto stim relative to start of
    sync. There should be no nans: the times will be used as trial start_time."""
    _stim_offset_times: npt.NDArray[np.float64]

    @functools.cached_property
    def start_time(self) -> npt.NDArray[np.float64]:
        with contextlib.suppress(AttributeError):
            return self._stim_onset_times
        return utils.safe_index(self._frame_times, self._hdf5["trialOptoOnsetFrame"][:])

    @functools.cached_property
    def stop_time(self) -> npt.NDArray[np.float64]:
        with contextlib.suppress(AttributeError):
            return self._stim_offset_times
        return self.start_time + self._hdf5["trialOptoDur"][:]

    @functools.cached_property
    def trial_index(self) -> npt.NDArray[np.int32]:
        """0-indexed"""
        return np.arange(len(self.start_time))

    @functools.cached_property
    def stim_id(self) -> npt.NDArray[np.int32]:
        return np.full_like(self.start_time, np.nan)



from __future__ import annotations

import functools
from collections.abc import Iterable, Iterator
from typing import ClassVar, Protocol

import npc_lims
import pandas as pd
import polars as pl
import pynwb


class SupportsToNWB(Protocol):
    def to_nwb(self, nwb: pynwb.NWBFile) -> None:
        ...


class NWBContainer(SupportsToNWB):
    add_to_nwb_method: ClassVar[str] = NotImplemented

    records: tuple[npc_lims.Record, ...]

    def __contains__(self, key: npc_lims.Record) -> bool:
        return key in self.records

    def __iter__(self) -> Iterator[npc_lims.Record]:
        return iter(self.records)

    def __len__(self) -> int:
        return len(self.records)

    def __init__(self, records: Iterable[npc_lims.Record], **kwargs) -> None:
        self.records = tuple(records)
        for key, value in kwargs.items():
            setattr(self, key, value)

    def to_nwb(self, nwb: pynwb.NWBFile) -> None:
        for record in self.records:
            getattr(nwb, self.add_to_nwb_method)(**record.__dict__)


class NWBContainerWithDF(NWBContainer):
    def to_dataframe(self) -> pd.DataFrame:
        return self.df.to_pandas()

    @functools.cached_property
    def df(self) -> pl.DataFrame:
        return pl.from_records(self.records)


class Epochs(NWBContainerWithDF):
    records: tuple[npc_lims.Epoch, ...]
    add_to_nwb_method = "add_epoch"


class Devices(NWBContainerWithDF):
    records: tuple[npc_lims.Device, ...]
    add_to_nwb_method = "add_device"


class ElectrodeGroups(NWBContainerWithDF):
    records: tuple[npc_lims.ElectrodeGroup, ...]
    add_to_nwb_method = "add_electrode_group"


class Electrodes(NWBContainerWithDF):
    records: tuple[npc_lims.Electrode, ...]
    add_to_nwb_method = "add_electrode"


class Units(NWBContainerWithDF):
    records: tuple[npc_lims.Units, ...]

    def add_to_nwb(self, nwb: pynwb.NWBFile) -> None:
        nwb.units = pynwb.misc.Units.from_dataframe(self.to_dataframe(), name="units")


class Intervals(NWBContainerWithDF):
    """Pass `name`, `description` and `column_names_to_descriptions` as kwargs."""

    records: tuple[npc_lims.Epoch, ...]
    name: str
    description: str
    column_names_to_descriptions: dict[str, str] = {}

    def add_to_nwb(self, nwb: pynwb.NWBFile) -> None:
        module = pynwb.epoch.TimeIntervals(
            name=self.name,
            description=self.description,
        )
        for key in self.records[0].__dict__.keys():
            if key in ("start_time", "stop_time"):
                continue
            module.add_column(
                name=key, description=self.column_names_to_descriptions.get(key, "")
            )

        for record in self.records:
            module.add_row(**record.__dict__)

        nwb.add_time_intervals(module)



"""
>>> path = upath.UPath('s3://aind-ephys-data/ecephys_668759_2023-07-11_13-07-32/ecephys_clipped/Record Node 102/settings.xml')
>>> et = ET.parse(io.BytesIO(path.read_bytes()))
>>> hostname(et)
'W10DT05516'
>>> date_time(et)
(datetime.date(2023, 7, 11), datetime.time(13, 7, 53))
>>> open_ephys_version(et)
'0.6.4'
>>> settings_xml_md5(path)
'5c1b33293cb7c5f72df56fbdd4b72fc3'
>>> probe_serial_numbers(et)
(18194810652, 18005123131, 18005102491, 19192719021, 18005118602, 19192719061)
>>> probe_letters(et)
('A', 'B', 'C', 'D', 'E', 'F')
>>> probe_types(et)
('Neuropixels 1.0', 'Neuropixels 1.0', 'Neuropixels 1.0', 'Neuropixels 1.0', 'Neuropixels 1.0', 'Neuropixels 1.0')
>>> isinstance(settings_xml_info_from_path(path), SettingsXmlInfo)
True
"""

from __future__ import annotations

import dataclasses
import datetime
import doctest
import functools
import hashlib
import io
import xml.etree.ElementTree as ET
from typing import Literal

import upath


@dataclasses.dataclass
class SettingsXmlInfo:
    """Info from a settings.xml file from an Open Ephys recording."""

    path: upath.UPath
    probe_serial_numbers: tuple[int, ...]
    probe_types: tuple[str, ...]
    probe_letters: Literal["A", "B", "C", "D", "E", "F"]
    hostname: str
    date: datetime.date
    start_time: datetime.time
    open_ephys_version: str


def settings_xml_info_from_path(path: str | upath.UPath) -> SettingsXmlInfo:
    """Info from a settings.xml file from an Open Ephys recording."""
    path = upath.UPath(path)
    et = ET.parse(io.BytesIO(path.read_bytes()))
    return SettingsXmlInfo(
        path=path,
        probe_serial_numbers=probe_serial_numbers(et),
        probe_types=probe_types(et),
        probe_letters=probe_letters(et),
        hostname=hostname(et),
        date=date_time(et)[0],
        start_time=date_time(et)[1],
        open_ephys_version=open_ephys_version(et),
    )


def get_tag_text(et: ET.ElementTree, tag: str) -> str | None:
    result = [
        element.text for element in et.getroot().iter() if element.tag == tag.upper()
    ]
    if not (result and any(result)):
        result = [element.attrib.get(tag.lower()) for element in et.getroot().iter()]
    return str(result[0]) if (result and any(result)) else None


def get_tag_attrib(et: ET.ElementTree, tag: str, attrib: str) -> str | None:
    result = [
        element.attrib.get(attrib)
        for element in et.getroot().iter()
        if element.tag == tag.upper()
    ]
    return str(result[0]) if (result and any(result)) else None


def hostname(et: ET.ElementTree) -> str:
    result = (
        # older, pre-0.6.x:
        get_tag_text(et, "machine")
        # newer, 0.6.x:
        or get_tag_attrib(et, "MACHINE", "name")
    )
    if not result:
        raise LookupError(f"No hostname: {result!r}")
    return result


@functools.cache
def date_time(et: ET.ElementTree) -> tuple[datetime.date, datetime.time]:
    """Date and recording start time."""
    result = get_tag_text(et, "date")
    if not result:
        raise LookupError(f"No datetime found: {result!r}")
    dt = datetime.datetime.strptime(result, "%d %b %Y %H:%M:%S")
    return dt.date(), dt.time()


@functools.cache
def probe_attrib_dicts(et: ET.ElementTree) -> tuple[dict[str, str], ...]:
    return tuple(
        probe_dict.attrib
        for probe_dict in et.getroot().iter()
        if "probe_serial_number" in probe_dict.attrib
    )


def probe_attrib(et: ET.ElementTree, attrib: str) -> tuple[str, ...]:
    return tuple(probe[attrib] for probe in probe_attrib_dicts(et))


def probe_serial_numbers(et: ET.ElementTree) -> tuple[int, ...]:
    return tuple(int(_) for _ in probe_attrib(et, "probe_serial_number"))


def probe_types(et: ET.ElementTree) -> tuple[str, ...]:
    try:
        return probe_attrib(et, "probe_name")
    except KeyError:
        return tuple("unknown" for _ in probe_attrib_dicts(et))


def probe_idx(et: ET.ElementTree) -> tuple[int, ...]:
    """Try to reconstruct index from probe slot and port.

    Normally 2 slots: each with 3 ports in use.
    """
    slots, ports = probe_attrib(et, "slot"), probe_attrib(et, "port")
    result = tuple(
        (int(s) - int(min(slots))) * len(set(ports)) + int(p) - 1
        for s, p in zip(slots, ports)
    )
    if not all(idx in range(6) for idx in result):
        raise ValueError(f"probe_idx: {result!r}, slots: {slots}, ports: {ports}")
    return result


def probe_letters(et: ET.ElementTree) -> Literal["A", "B", "C", "D", "E", "F"]:
    return tuple("ABCDEF"[idx] for idx in probe_idx(et))  # type: ignore


def open_ephys_version(et: ET.ElementTree) -> str:
    result = get_tag_text(et, "version")
    if not result:
        raise LookupError(f"No version found: {result!r}")
    return result


def settings_xml_md5(path: str | upath.UPath) -> str:
    return hashlib.md5(upath.UPath(path).read_bytes()).hexdigest()


if __name__ == "__main__":
    import doctest

    import dotenv

    dotenv.load_dotenv(dotenv.find_dotenv(usecwd=True))
    doctest.testmod(
        optionflags=(doctest.IGNORE_EXCEPTION_DETAIL | doctest.NORMALIZE_WHITESPACE)
    )
